//======================================================== IBJplume.c
//
// IBJplume: Gaussian plume model
// ==============================
//
// Copyright (C) Janicke Consulting, Dunum, Germany, 2001
// email: info@janicke.de
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// history:
// 2001-07-30 uj 1.1.0 extracted from gaufah 1.5.17 
// 2001-08-15 uj 1.2.0 release candidate
// 2001-08-27 uj 1.2.1 bug in ClrPlume() corrected
// 2001-08-27 uj 1.2.2 log information changed in IBJstt.c
// 2001-09-11 uj 1.2.3 re-definition of aq, bq, rq
//
//==================================================================
#define  VERSION  "1.2.3"

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <math.h> 
#include <time.h>
#include <process.h>

#include "IBJdmn.h"
#include "IBJary.h"
#include "IBJtxt.h"
#include "IBJmsg.h"
#include "IBJmth.h"
#include "IBJtbl.h"
#include "IBJqtl.h"
#include "IBJplume.h"

static char *eMODn = "IBJplume";

static ARYDSC ConDsc;
static int doCon;
#define CON(a,b)    (*(float*)AryPtrX(&ConDsc,(a),(b)))
static ARYDSC DryDsc;
static int doDry;
#define DRY(a,b)    (*(float*)AryPtrX(&DryDsc,(a),(b)))
static ARYDSC WetDsc;
static int doWet;
#define WET(a,b)    (*(float*)AryPtrX(&WetDsc,(a),(b)))
static ARYDSC TrnDsc;
static int doTrn;
#define TRN(a,b)    (*(float*)AryPtrX(&TrnDsc,(a),(b)))
static int doQtl;


/* ------------------------------------------------------------------------ */
  
static struct { short i, j; float v, w; } EwtRec;
static int Md, Verbose, Newt, Initialized=0;
static float DetMin, ArgMax, SigzMax, F1, Dauer;
static int Ng[2];
static float Xg[2][12], Wg[2][12];
static float **Cc, **Aa, **Sy;
static float **Cz, **Ch;
static float **Mm, **Uu;
static float *Cona, *Drya, *Weta, *Trna;
static FILE *LogFile, *EwtFile;
static char LogName[] = "plmlog.txt";
static char CmdName[] = "plmcmd.txt"; 
static char Title[80];
static float TstZa, TstDz;
static int TstNi;
static char WorkPath[MAXFNLEN] = "./";

static float Ew1[NKL] = { 0.37, 0.32, 0.26, 0.18, 0.14, 0.12 };  /* glattes Gelnde */
static float Ew2[NKL] = { 0.52, 0.48, 0.31, 0.31, 0.31, 0.20 };  /* rauhes Gelnde  */
static float Ew3[NKL] = { 0.42, 0.37, 0.28, 0.22, 0.20, 0.09 };  /* TA-Luft         */
static float Hm1[NKL] = {  250,  250,  800,  800, 1100, 1100 };  /* VDI 3782/1 */

int Nsig = 3;
static float HsigNEW[3] = { 50, 100, 180 };
static struct { float F, f, G, g; } SigTabNEW[3][NKL] =
  {{{ 1.294, 0.718, 0.241, 0.662 },
    { 0.801, 0.754, 0.264, 0.774 },
    { 0.640, 0.784, 0.215, 0.885 },
    { 0.659, 0.807, 0.165, 0.996 },
    { 0.876, 0.823, 0.127, 1.108 },
    { 1.503, 0.833, 0.151, 1.219 }},
   {{ 0.253, 1.057, 0.717, 0.486 },
    { 0.411, 0.882, 0.487, 0.652 },
    { 0.504, 0.818, 0.265, 0.818 },
    { 0.466, 0.866, 0.137, 0.985 },
    { 0.324, 1.025, 0.070, 1.151 },
    { 0.170, 1.296, 0.051, 1.317 }},
   {{ 0.671,  0.903,  0.484,  0.500  },
    { 0.345,  0.903,  0.546,  0.557  },
    { 0.208,  0.903,  0.307,  0.734  },
    { 0.232,  0.903,  0.104,  0.997  },
    { 0.415,  0.903,  0.0330, 1.320  },
    { 0.671,  0.903,  0.0245, 1.500  }}};

static float HsigTAL[3] = { 50, 100, 150 };    
static struct { float F, f, G, g; } SigTabTAL[3][NKL] =
    {{{ 1.294, 0.718, 0.241, 0.662 },
    { 0.801, 0.754, 0.264, 0.774 },
    { 0.640, 0.784, 0.215, 0.885 },
    { 0.659, 0.807, 0.165, 0.996 },
    { 0.876, 0.823, 0.127, 1.108 },
    { 1.503, 0.833, 0.151, 1.219 }},
     {{ 0.253, 1.057, 0.717, 0.486 },
    { 0.411, 0.882, 0.487, 0.652 },
    { 0.504, 0.818, 0.265, 0.818 },
    { 0.466, 0.866, 0.137, 0.985 },
    { 0.324, 1.025, 0.070, 1.151 },
    { 0.170, 1.296, 0.051, 1.317 }},
     {{ 0.31,  0.71,  0.06,  0.71  },
    { 0.31,  0.71,  0.06,  0.71  },
    { 0.32,  0.78,  0.22,  0.78  },
    { 0.36,  0.86,  0.33,  0.86  },
    { 0.40,  0.91,  0.41,  0.91  },
    { 0.40,  0.91,  0.41,  0.91  }}};

    
/*============================== Prototypes ===============================*/

static char *MakeFileName( char *name, char *extension )
  ;
static int Init( int level );    /* Initialisierung der Variablen */

static int GetG( void );         /* Parameter fr Gau-Quadratur  */

static int Sigma(  /* Berechnung von Sigma-Z und, falls gewnscht, Sigma-Y. */
float x,           /* Abstand von der Quelle                                */
float z,           /* Hhe ber dem Erdboden                                */
float *sz,         /* Adresse von Sigma-Z                                   */
int sigmay,        /* != 0, falls Sigma-Y berechnet werden soll.            */
float *sy );       /* Adresse von Sigma-Y                                   */

static void Teile(  /* Berechnung der zur Konzentration beitragenden Teile */
float z,            /* Hhe ber dem Erdboden                              */
float ha,           /* Hhe der Fahnen-Achse ohne Sedimentation            */
float hg,           /* Absenkung durch Sedimentation                       */
float sz,           /* Sigma-Z                                             */
float mh,           /* Hhe der Mischungsschicht                           */
float *gp,          /* Anteil der realen Quelle (mit Spiegelbildern)       */
float *gm,          /* Anteil der virtuellen Quelle (mit Spiegelbildern)   */
int fluss,          /* != 0, falls die Flsse berechnet werden sollen      */
float *fp,          /* Flu von der realen Quelle (mit Spiegelbildern)     */
float *fm);         /* Flu von der virtuellen Quelle (mit Spiegelbildern) */

static int Tilted(  /* Berechnung der Fahnen-Funktionen nach Tilted Plume */
int iq );           /* Index der Quelle                                   */


static void Uprofil(    /* Berechnung des Windprofils    */
float z,                /* Hhe ber dem Erdboden        */
float *pu);             /* Pointer auf Geschwindigkeit   */

static void Wind(     /* Berechnung der mittleren Transportgeschwindigkeit */
float ha,             /* Hhe der Fahnen Achse ohne Sedimentation          */
float hg,             /* Absenkung der Fahnenachse durch Sedimentation     */
float sz,             /* Sigma-Z                                           */
float af,      /* Faktor fr Anteil der Spiegelfahne                */
float *pu );   /* Pointer auf Windgeschwindigkeit                   */

static int ClrPlume( /* Fahnen-Funktionen lschen             */
int level );         /* =0: nur freigeben,  =1: neu anlegen.  */

static int PrnPlume(    /* Fahnen-Funktionen im Log-File tabellieren */
int iq );               /* Index der Quelle                          */

static int Boden(   /* Berechnung der bodennahen Werte   */
int iq,             /* Index der Quelle                  */
float xa,           /* X-Koordinate des Aufpunktes       */
float ya,           /* Y-Koordinate des Aufpunktes       */
float si,           /* Sinus des Winkels Fahne/X-Achse   */
float co,           /* Kosinus des Winkels Fahne/X-Achse */
float *pwk,         /* Adresse Konzentrationswert (z=ZP) */
float *pwh,         /* Adresse Konzentrationswert (z=HA) */
float *pwn,         /* Adresse nasse Deposition          */
float *pwu );       /* Adresse umgewandelter Anteil      */

static int ClcCnc(  /* Berechnung der Konzentration      */
int iq,             /* Index der Quelle                  */
float xa,           /* X-Koordinate des Aufpunktes       */
float ya,           /* Y-Koordinate des Aufpunktes       */
float si,           /* Sinus des Winkels Fahne/X-Achse   */
float co,           /* Kosinus des Winkels Fahne/X-Achse */
float *pwk,         /* Adresse Konzentrationswert        */
float *pwi,         /* Intermittenz                      */
float *pwn,         /* Adresse nasse Deposition          */
float *pwu )        /* Adresse umgewandelter Anteil      */
  ;
static int DefRaster( void );  /* Platz fr Rasterwerte anlegen  */

static int ClcRaster( void );  /* Werte auf dem Raster berechnen */

static int WriteResult( char *arg )     // Rasterausgabe einleiten
;

      
/*============================================================== MakeFileName
*/
static char *MakeFileName( char *name, char *extension )
  {
  dP(MakeFileName);
  static char fname[MAXFNLEN+1];
  char *pc, *ps, *pe, n[MAXFNLEN+1];
  if (strlen(name) > MAXFNLEN)                          eX(1);
  strcpy(n, name);
  for (pc=n; (*pc); pc++)  if (*pc == ALTPSEP)  *pc = PATHSEP;
  ps = strrchr(n, PATHSEP);
  pe = strrchr(n, '.');
  if ((ps) && (pe) && (pe < ps))  pe = NULL;
  if (ps)  *fname = 0;
  else  strcpy(fname, WorkPath);
  if (strlen(fname) + strlen(n) > MAXFNLEN)             eX(2);
  strcat(fname, n);
  if (!pe) {
    if (strlen(fname) + strlen(extension) > MAXFNLEN)   eX(3);
    strcat(fname, extension);
    }
  return fname;
eX_1:  eX_2:  eX_3:
  *fname = 0;
  nMSG("*** File-Name zu lang! (%s)\n", ePGMs);
  return fname;
}

//===================================================================== Stat
int Stat( void )
{
  dP(Stat);
  int iq, rc;
  ClrPlume(1);                       eG(4);
  for (iq=0; iq<NQ; iq++) {
    rc = Tilted(iq); if (rc < 0)     eG(6);
  }
  rc = ClcRaster(); if (rc < 0)      eG(7);
  return 0;
eX_4:
  eMSG("*** can't initialize plume functions! (%s)\n", ePGMs);
eX_6:
  eMSG("*** can't calculate plume functions! (%s)\n", ePGMs);
eX_7:
  eMSG("*** can't calculate raster values! (%s)\n", ePGMs);
}

/*====================================================================== Init
*/  

static int Init( int level )    /* Initialisierung der Variablen */
  {
  dP(Init);
  F1 = 1.0/sqrt(2.*3.14159265359);
  if (strstr(OP, "taluft"))    Md |= MO_TALUFT;
  //if (strstr(OP, "berljand"))  Md |= MO_BERLJAND;
  doCon = 1;
  doDry = 1;  
  doWet = 1;  
  doTrn = 1; 
  doQtl = 1;
  if (NH<1 && !MODUS(BERLJAND))                         eX(104);
  if (NQ < 1)                                           eX(105);
  if (NW < 1)                                           eX(106);
  if (NULL == (Cc = CALLOC(NQ, sizeof(float *))))       eX(121);
  if (NULL == (Cz = CALLOC(NQ, sizeof(float *))))       eX(122);
  if (NULL == (Ch = CALLOC(NQ, sizeof(float *))))       eX(123);
  if (NULL == (Aa = CALLOC(NQ, sizeof(float *))))       eX(124);
  if (NULL == (Sy = CALLOC(NQ, sizeof(float *))))       eX(125);
  if (NULL == (Mm = CALLOC(NQ, sizeof(float *))))       eX(126);
  if (NULL == (Uu = CALLOC(NQ, sizeof(float *))))       eX(127);
  if (NP > 0) {
    if (NULL == (Cona = CALLOC(NP, sizeof(float))))     eX(130);
    if (NULL == (Drya = CALLOC(NP, sizeof(float))))     eX(130);
    if (NULL == (Weta = CALLOC(NP, sizeof(float))))     eX(130);
    if (NULL == (Trna = CALLOC(NP, sizeof(float))))     eX(130);
  }
  Initialized = 1;
  return 0; 
eX_104:
  eMSG("*** Ungltige Anzahl (=%d) von Hhenstufen! (%s)\n", NH, ePGMs);
eX_105:
  eMSG("*** Ungltige Anzahl (=%d) von Quellen! (%s)\n", NQ, ePGMs);
eX_106:
  eMSG("*** Ungltige Anzahl (=%d) von Windsektoren! (%s)\n", NW, ePGMs);
eX_121: eX_122: eX_123: eX_124: eX_125: eX_126: eX_127: eX_130:
  eMSG("*** Interne Tabelle kann nicht angelegt werden! (%s)\n", ePGMs);
}
 
/* ====================================================================== GetG
 *
 *     Zwei Stze von Gewichten und Abszissen fr die Gau-Quadratur     
 *     werden in Xg[][] und Wg[][] bereitgestellt. Sie dienen zur Be-    
 *     rechnung von Integralen des Typs                                  
 *                                                                       
 *       Integral von 0 bis 1 ber f(x) dx                               
 * -------------------------------------------------------------------------*/

static int GetG( void )     /* Parameter fr Gau-Quadratur  */
  {
  /*     nn     : Halbe Anzahl der Sttzpunkte
   *     xx, ww : Positive Abszissen und zugehrige Gewichte fr normale
   *              Gau-Quadratur auf dem Intervall [-1,+1]  
   *              (aus Abramowitz & Stegun).
   */
  dP(GetG);
  static int nn[] = { 1, 2, 3, 4, 6 };
  static double xx[5][6] = {
             { 0.577350269189626e0, 0, 0, 0, 0, 0 },
             { 0.339981043584856e0,
               0.861136311594053e0, 0, 0, 0, 0 },
             { 0.238619186083197e0,
               0.661209386466265e0,
               0.932469514203152e0, 0, 0, 0 },
             { 0.183434642495650e0,
               0.525532409916329e0,
               0.796666477413627e0,
               0.960289856497536e0, 0, 0 },
             { 0.125233408511469e0,
               0.367831498998180e0,
               0.587317954286617e0,
               0.769902674194305e0,
               0.904117256370475e0,
               0.981560634246719e0 } };

  static double ww[5][6] = {
             { 1.000000000000000e0, 0, 0, 0, 0, 0 },
             { 0.652145154862546e0,
               0.347854845137454e0, 0, 0, 0, 0 },
             { 0.467913934572691e0,
               0.360761573048139e0,
               0.171324492379170e0, 0, 0, 0 },
             { 0.362683783378362e0,
               0.313706645877887e0,
               0.222381034453374e0,
               0.101228536290376e0, 0, 0 },
             { 0.249147045813403e0,
               0.233492536538355e0,
               0.203167426723066e0,
               0.160078328543346e0,
               0.106939325995318e0,
               0.047175336386512e0 } };
  int j, k, l, n;

  /*   --------------- Umrechnen auf das Intervall [0,1] ---------------- */

  for (l=0; l<2; l++) {
    if (l == 0)  n = N1;
    else  n = N2;
    for (j=0; j<5; j++)
      if (2*nn[j] >= n)  break;
    if (j >= 5)                         eX(111);
    n = nn[j];
    Ng[l] = 2*n;
    for (k=0; k<n; k++) {
      Xg[l][n+k]   = 0.5*(1.+xx[j][k]);
      Xg[l][n-1-k] = 0.5*(1.-xx[j][k]);
      Wg[l][n+k]   = 0.5*ww[j][k];
      Wg[l][n-1-k] = 0.5*ww[j][k]; }
    }
  return 0;
eX_111:
  eMSG("*** Ungltige Anzahl (=%d) von Sttzpunkten! (%s)\n", n, ePGMs);
}
  
//=============================================================== Help
static void Help( void )
{
  MsgVerbose = 1;
  vMsg("\nIBJplume: Gaussian plume model, version %s\n", VERSION);
  vMsg("Copyright (C) Janicke Consulting, Dunum, Germany, 2001\n");
  vMsg("IBJplume comes with ABSOLUTELY NO WARRANTY\n");
  vMsg("IBJplume is free software under the GNU PUBLIC LICENCE\n");
  vMsg("Compiled at %s  %s\n", __DATE__, __TIME__);
  PrintHelp();
  vMsg("%s",
   "\nParameter drfen nur in den zugehrigen Abschnitten gesetzt werden.\n"
   "In der folgenden Tabelle sind alle Parameter mit Namen (zwei Buchstaben,\n"
   "Gro-/Kleinschreibung ist nicht signifikant), Standardwert, Vielfachheit\n"
   "(bei Vektoren) und ihrer Bedeutung aufgelistet: \n\n");
  PrmPrint("*c"); vMsg("\n");
  PrmPrint("*d"); vMsg("\n");
  PrmPrint("*g"); vMsg("\n");
  PrmPrint("*l"); vMsg("\n");
  PrmPrint("*p"); vMsg("\n");
  PrmPrint("*q"); vMsg("\n");
  PrmPrint("*s"); vMsg("\n");
  PrmPrint("*t"); vMsg("\n");
  PrmPrint("*w"); vMsg("\n");
  PrmPrint("*e"); vMsg("\n");
  exit(0);
}

/*================================================================ PrintHelp
*/
static void PrintHelp( void )   
  {
  vMsg("%s",
  "\nAufruf: IBJplume Arbeitsverzeichnis [Optionen]\n\n"
  "Im Arbeitsverzeichnis stehen die Eingabe- und Ausgabedateien.\n"
  "Ihre Namen sind:\n"
  "  plmcmd.txt    Eingabedatei,\n"
  "  plmlog.txt    Log-Datei zur Protokollierung des Ablaufs,\n"
  "Bei dem Kommando *s <Option> zustzlich:\n"
  "  cnc.dmna      Gitterwerte der Konzentration (Option cnc),\n"
  "  dry.dmna      Gitterwerte der trockenen Deposition (Option dry),\n"
  "  wet.dmna      Gitterwerte der nassen Deposition (Option wet),\n"
  "  trn.dmna      Gitterwerte der umgesetzten Konzentration (Option trn),\n\n");
  vMsg("%s",
  "Auflistung der Kommandos. Bei den Abschnitten ist nur der erste Buchstabe\n"
  "signifikant (*c = *CLEAR), bei den Kommandos nur die ersten zwei Zeichen:\n\n"
  "*CLEAR\n"
  "   Werte auf dem Gitter und den Aufpunkten lschen.\n\n"
  "*DIMENSIONIERUNG (mu als erstes angegeben werden)\n"
  "   Festlegung von Dimensionsgrenzen einiger interner Tabellen.\n\n"
  "*GITTER (darf nur einmal angegeben werden)\n"
  "   Festlegung eines Rasters von Aufpunkten, an denen bodennahe Konzen-\n"
  "   tration, trockene oder nasse Deposition oder umgewandelte Konzentra-\n"
  "   tion berechnet werden sollen.\n\n");
  vMsg("%s",
  "*LISTEN plm|pnt\n"
  "   Bei plm:\n"
  "   Tabellierung der aktuellen Fahnenfunktionen im Log-File fr alle Quellen.\n"
  "   Es bedeuten: Iy       : Normierte Konzentration ber y integriert\n"
  "                Iyz      : Normierte Konzentration ber y und z integriert\n"
  "                Umwandl. : Umgewandelter Anteil in der Konzentration\n"
  "                Sigma-Y  : Horizontale Breite der Fahne\n"
  "                Konzentr.: Konzentration am Boden auf der Fahnenachse\n"
  "   Die normierte Konzentration ist auf die Quellstrke 1 ME/s bezogen.\n\n"
  "   Bei pnt:\n"
  "   Auflistung der an den Aufpunkten berechneten Werte in der Log-Datei.\n"
  "   Es bedeuten: Cm : Konzentrationswert\n"
  "                Fd : Trockene Deposition\n"
  "                Fw : Nasse Deposition\n"
  "                Tr : Umgewandelte Konzentration\n");
  vMsg("%s",
  "*PARAMETER\n"
  "   Festlegung verschiedener Parameter (z.B. Meteorologie, Ausbreitungs-\n"
  "   parameter, Schrittweiten) fr die Ausbreitungsrechnung.\n"
  "   Berechnung der Fahnen: Wenn ein Gitter oder Aufpunkte definiert\n"
  "   sind, werden die Werte an den Punkten berechnet und zu den bereits\n"
  "   bereits gespeicherten Werte addiert (gem der Dauer der aktuellen\n"
  "   Windrichtung gewichtet).\n\n"
  "*QUELLEN\n"
  "   Festlegung von Lage und Ausdehnung der Quellen, der Quellstrken \n"
  "   und der berhhungen.\n\n"
  "*SPEICHERN con|dry|wet|trn\n"
  "   Ausgabe des betreffenden Feldes auf den Gitterpunkten. Die ausgegebenen\n"
  "   Werte sind Mittelwerte ber alle vorher berechneten Fahnen.\n\n");
  vMsg("%s",
  "*TABELLE * oder c|d|g|l|p|q|s|w\n"
  "   Ausgabe der aktuellen Parameterwerte. Dabei knnen durch Angabe von\n"
  "   Kennbuchstaben bestimmte Abschnitte ausgewhlt werden.\n\n"
  "*WETTERSTATISTIK <datei>\n"
  "   Die Ausbreitungsklassenstatistik <datei> (DWD-Format) wird eingelesen\n"
  "   und automatisch abgearbeitet.\n\n"
  "*ENDE (mu als letztes angegeben werden)\n"
  "   Beenden des Programms.\n");
  return; 
}

/* ===================================================================== Sigma
*/                                                                       
static int Sigma(  /* Berechnung von Sigma-Z und, falls gewnscht, Sigma-Y. */
float x,           /* Abstand von der Quelle                                */
float z,           /* Hhe ber dem Erdboden                                */
float *sz,         /* Adresse von Sigma-Z                                   */
int sigmay,        /* != 0, falls Sigma-Y berechnet werden soll.            */
float *sy )        /* Adresse von Sigma-Y                                   */
  {      
  float a1, a2, lggi, kgi, lgfi, kfi, lx;
  int n;
  if (x < 1) {            /* Minimaler Quellabstand ist 1 m */
    *sz = 0;
    if (sigmay) *sy = 0;
    return 0; 
  }
  for (n=0; n<NH; n++) 
    if (HS(n) >= z)  break;
  if ((n == 0) || (n == NH)) {
    if (n == NH)  n--;
    *sz = PZ(n)*pow(x,QZ(n));
    if (sigmay)  *sy = PY(n)*pow(x,QY(n)); }
  else {
    a1 = (z-HS(n-1))/(HS(n)-HS(n-1));
    a2 = 1.-a1;
    lx = log(x);
    lggi = a2*log(PZ(n-1)) + a1*log(PZ(n));
    kgi  = a2*QZ(n-1) + a1*QZ(n);
    *sz  = exp(lggi + kgi*lx);
    if (sigmay) {
      lgfi = a2*log(PY(n-1)) + a1*log(PY(n));
      kfi  = a2*QY(n-1) + a1*QY(n);
      *sy  = exp(lgfi + kfi*lx); }
    }
  if (*sz > SigzMax)  *sz = SigzMax;
  return 0; }      

/* ===================================================================== Teile
*/
static void Teile(  /* Berechnung der zur Konzentration beitragenden Teile */
float z,     /* Hhe ber dem Erdboden                              */
float ha,    /* Hhe der Fahnen-Achse ohne Sedimentation            */
float hg,    /* Absenkung durch Sedimentation                       */
float sz,    /* Sigma-Z                                             */
float mh,    /* Hhe der Mischungsschicht                           */
float *gp,   /* Anteil der realen Quelle (mit Spiegelbildern)       */
float *gm,   /* Anteil der virtuellen Quelle (mit Spiegelbildern)   */
int fluss,   /* != 0, falls die Flsse berechnet werden sollen      */
float *fp,   /* Flu von der realen Quelle (mit Spiegelbildern)     */
float *fm)   /* Flu von der virtuellen Quelle (mit Spiegelbildern) */
  {
  float a, w2, rw2pi, hp, hm, ap, am, d, bp, bm, hpi, hmi, w2sz;
  int i, n;
  w2 = 1.414213562;
  rw2pi = 0.39894228;
  w2sz = w2*sz;
  d = 2*mh;
  hp =  ha - hg;
  hm = -ha - hg;
  n = (int)(0.5+(MS*sz+hp)/d);
  *gp = 0;
  *gm = 0;
  if (fluss) {
    *fp = 0;
    *fm = 0; }
  if (z <= 0) {
    ap = -hp/(w2*sz);
    am = -hm/(w2*sz);
    if (fluss) {
      *fp = 0.5*(float)Erfc(ap);
      *fm = 0.5*(float)Erfc(am); }
    a = ap*ap;
    if (a < ArgMax)  *gp = exp(-a);
    for (i=1; i<=n; i++) {
      ap = -(hp - i*d)/(w2*sz);
      a = ap*ap;
      if (a < ArgMax)  *gp += 2*exp(-a);
      }
    *gp *= rw2pi/sz;
     a = am*am;
    if (a < ArgMax)  *gm = exp(-a);
    for (i=1; i<=n; i++) {
      am = -(hm - i*d)/(w2*sz);
      a = am*am;
      if (a < ArgMax)  *gm += 2*exp(-a);
      }
    *gm *= rw2pi/sz;
    return; }
  for (i=-n; i<=n; i++) {
    if (i>= 0) {
      hpi = hp - i*d;
      hmi = hm - i*d; }
    else {
      hpi = -hp - i*d;
      hmi = -hm - i*d; }
    ap = (z - hpi)/w2sz;
    a = ap*ap;
    if (a < ArgMax)  *gp += exp(-a);
    am = (z - hmi)/w2sz;
    a = am*am;
    if (a < ArgMax)  *gm += exp(-a);
    if (fluss) {
      bp = (mh - hpi)/w2sz;
      bm = (mh - hmi)/w2sz;
      *fp += 0.5*(float)(Erfc(ap) - Erfc(bp));
      *fm += 0.5*(float)(Erfc(am) - Erfc(bm)); }
    }
  
  *gp *= rw2pi/sz;
  *gm *= rw2pi/sz;
/* printf("TEILE: gp=%10.2e, gm=%10.2e, fp=%10.2e, fm=%10.2e\n",
*gp,*gm,*fp,*fm);    */
  return; }

/*==================================================================== Uprofil  
*/
static void Uprofil(    /* Berechnung des Windprofils    */
float z,                /* Hhe ber dem Erdboden        */
float *pu)              /* Pointer auf Geschwindigkeit   */
{
  if (z > HW)  z = HW;
  if (z <= 0)  *pu = 0;
  else {
    if (EW <= 0)  *pu = UA;
    else  *pu = UA*pow(z/HA,EW); 
  }
  return; 
}

      
/*======================================================================= Wind
*/                                                                       
static void Wind(   /* Berechnung der mittleren Transportgeschwindigkeit */
float ha,           /* Hhe der Fahnen Achse ohne Sedimentation          */
float hg,           /* Absenkung der Fahnenachse durch Sedimentation     */
float sz,           /* Sigma-Z                                           */
float af,           /* Faktor fr Anteil der Spiegelfahne                */
float *pu )         /* Pointer auf Windgeschwindigkeit                   */
{      
  float u, g, z, zg[4], z1, z2, dz, hp;
  float up, ut, gt, c, gp, gm;
  float *xg, *wg;
  int j, k, ng;
  if MODUS(TALUFT)  sz = 0;        /* TA-Luft Modus */
  if (sz <= 0) {
    Uprofil( ha-hg, pu );
    return; 
  }
  hp =  ha - hg;                   /*-23mar94-*/
  if (hp < 0)  hp = 0;             /*-28mar94-*/
  zg[1] = hp - 2*sz;
  if (zg[1] < 0)  zg[1] = 0;
  zg[0] = zg[1];
  zg[2] = hp + 2*sz;
  if (zg[2] > HM)  zg[2] = HM;
  zg[3] = zg[2];
  if ((HW < zg[3]) && (HW > zg[1]))  zg[2] = HW;
  if ((HA < zg[2]) && (HA > zg[0]))  zg[1] = HA;
  if (zg[0] >= zg[3]) {
    printf("*** Fehler in der Berechnung des mittleren Windes!\n");
    printf("zg = { %1.1f  %1.1f  %1.1f  %1.1f }\n",
      zg[0], zg[1], zg[2], zg[3]);
    exit(1); 
  }
  u = 0;  g = 0;
  ng = Ng[1];
  xg = Xg[1];
  wg = Wg[1];
  for (k=0; k<3; k++) {
    z1 = zg[k];
    z2 = zg[k+1];
    if (z2 <= z1)  continue;
    dz = z2 - z1;
    ut = 0;  gt = 0;
    for (j=0; j<ng; j++) {
      z = z1 + dz*xg[j];
      Uprofil( z, &up );
      Teile( z, ha, hg, sz, HM, &gp, &gm, FALSE, NULL, NULL );
      c = gp + af*gm;
      gt += wg[j]*c;
      ut += wg[j]*c*up;
    }
    g += dz*gt;
    u += dz*ut;
  }
  if (g > 0)  u /= g;
  else  u = 1.0;                   /*-28mar94-*/
  *pu = u;
  return; 
}

/*=================================================================== ClrPlume
*/
static int ClrPlume( /* Fahnen-Funktionen lschen             */
int level )          /* =0: nur freigeben,  =1: neu anlegen.  */
{ dP(ClrPlume);
  int i, nn, ikl, rc;
  if (Cc[0])  FREE(Cc[0]);
  for (i=0; i<NQ; i++)  Cc[i] = NULL;
  if (Cz[0])  FREE(Cz[0]);
  for (i=0; i<NQ; i++)  Cz[i] = NULL;
  if (Ch[0])  FREE(Ch[0]);
  for (i=0; i<NQ; i++)  Ch[i] = NULL;
  if (Aa[0])  FREE(Aa[0]);
  for (i=0; i<NQ; i++)  Aa[i] = NULL;
  if (Sy[0])  FREE(Sy[0]);
  for (i=0; i<NQ; i++)  Sy[i] = NULL;
  if (Mm[0])  FREE(Mm[0]);
  for (i=0; i<NQ; i++)  Mm[i] = NULL;
  if (Uu[0])  FREE(Uu[0]);
  for (i=0; i<NQ; i++)  Uu[i] = NULL;
  if (level == 0)  return 0;
  rc = GetG();
  if (rc < 0)  return rc;
  nn = NQ;
  for (i=0; i<NQ; i++)  nn += NI(i)+1;
  if (NULL == (Cc[0] = CALLOC( nn, sizeof(float) )))  return -301;
  if (NULL == (Cz[0] = CALLOC( nn, sizeof(float) )))  return -301;
  if (NULL == (Ch[0] = CALLOC( nn, sizeof(float) )))  return -301;
  if (NULL == (Aa[0] = CALLOC( nn, sizeof(float) )))  return -302;
  if (NULL == (Sy[0] = CALLOC( nn, sizeof(float) )))  return -303;
  if (NULL == (Mm[0] = CALLOC( nn, sizeof(float) )))  return -304;
  if (NULL == (Uu[0] = CALLOC( nn, sizeof(float) )))  return -305;
  for (i=1; i<NQ; i++) {   
    nn = NI(i-1)+1;
    Cc[i] = Cc[i-1] + nn;
    Cz[i] = Cz[i-1] + nn;
    Ch[i] = Ch[i-1] + nn;
    Aa[i] = Aa[i-1] + nn;
    Sy[i] = Sy[i-1] + nn; 
    Mm[i] = Mm[i-1] + nn; 
    Uu[i] = Uu[i-1] + nn; 
  }
  ikl = -1;
  if (PrmSet("kl")) {
    if (!strcmp(KL, "I"))          ikl = 0;
    else if (!strcmp(KL, "II"))    ikl = 1;
    else if (!strcmp(KL, "III/1")) ikl = 2;
    else if (!strcmp(KL, "III/2")) ikl = 3;
    else if (!strcmp(KL, "IV"))    ikl = 4;
    else if (!strcmp(KL, "V"))     ikl = 5;
    else ikl = 2;  
    if (!PrmSet("qy") && !PrmSet("py") && !PrmSet("qz") && !PrmSet("pz")) {
      NH = Nsig;
      if (!MODUS(TALUFT)) {
        for (i=0; i<NH; i++) {
          HS(i) = HsigNEW[i];
          PY(i) = SigTabNEW[i][ikl].F;
  						  QY(i) = SigTabNEW[i][ikl].f;
  						  PZ(i) = SigTabNEW[i][ikl].G;
          QZ(i) = SigTabNEW[i][ikl].g;
        }
      }
      else {
        for (i=0; i<NH; i++) {
          HS(i) = HsigTAL[i];
          PY(i) = SigTabTAL[i][ikl].F;
  						  QY(i) = SigTabTAL[i][ikl].f;
  						  PZ(i) = SigTabTAL[i][ikl].G;
          QZ(i) = SigTabTAL[i][ikl].g;
        }      
      }
    } 
    else                                                             eX(1);
    if (!PrmSet("ew")) EW = (RH == 0) ? Ew1[ikl] : Ew2[ikl];
    if (!PrmSet("hm")) HM = Hm1[ikl];
  }
  if (!PrmSet("hm") && MODUS(TALUFT) && ikl>-1) HM = 9999999.;
  if (!PrmSet("ew") && MODUS(TALUFT) && ikl>-1) EW = Ew3[ikl];
  return 0; 
eX_1: 
  eMSG("*** Es darf nicht gleichzeitig kl und py|qy|pz|qz angegeben werden! (%s)\n", 
  ePGMs);
}

/*=================================================================== PrnPlume
*/
static int PrnPlume(     /* Fahnen-Funktionen im Log-File tabellieren */
int iq )                 /* Index der Quelle                          */
  { 
  int i, ni;
  float *cz, *aa, *sy, *mm, *uu;
  float li, qq, fak;
  cz = Cz[iq];
  aa = Aa[iq];
  sy = Sy[iq];
  qq = EQ(iq);
  mm = Mm[iq];
  uu = Uu[iq];
  if (cz==NULL || aa==NULL || sy==NULL || mm==NULL || uu==NULL)  return -401;
  ni = NI(iq);
  li = LI(iq);
  fprintf( LogFile, "\nWerte bei z=%1.1f fr die Fahne von Quelle %d\n",
    ZP, iq );
  fprintf( LogFile, 
  "\nAbstand     Iy(Norm)    Iyz(Norm)  Umwandl.  Sigma-Y     Konzentr.\n");
  fprintf( LogFile,
  "    (m)      (%2s/m2)       (%2s/m)         -      (m)       (%1s/m3)\n\n", 
  ME, ME, ME);
  for (i=0; i<=ni; i+=(ND+1)) {
    if (sy[i] > 0)  fak = qq*F1/sy[i];
    else  fak = 0;
    fprintf( LogFile, "%7.0f %12.4e %12.4e %9.6f %8.1f %12.4e\n", 
                      i*li, cz[i], mm[i], uu[i], sy[i], cz[i]*fak); 
  }
  fprintf( LogFile, "\n" );
  return 0; 
}


static void TstVertikal( float x, float ha, float hg, float sz, float a )
  {
  float z, dz, gp, gm, fp, fm, uq;
  int i, ni;
  z = TstZa;
  dz = TstDz;
  ni = TstNi;
  if (dz <= 0)  return;
  Wind( ha, hg, sz, a, &uq );
  fprintf( LogFile, "\fTEST: x=%3.0f, ha=%3.1f, hg=%3.1f, sz=%3.1f, a=%4.3f\n",
           x, ha, hg, sz, a );
  fprintf( LogFile, 
  "    z          gp          gm          cc         fp         fm\n");
  for (i=0; i<ni; i++) {
    Teile( z, ha, hg, sz, HM, &gp, &gm, TRUE, &fp, &fm );
    fprintf( LogFile, "%5.0f %11.3e %11.3e %11.3e %11.3e %11.3e\n", 
      z, gp, gm, (gp+a*gm)/uq, fp, fm );
    z += dz;
    }
  fprintf( LogFile, "\n" );
  return; }

/* ==================================================================== Tilted
 *                                                                       
 *     Es wird die quer zur Windrichtung integrierte Bodenkonzentration cc *
 *     nach dem Tilted-Plume-Modell berechnet und zusammen mit dem Faktor  *
 *     der Spiegelfahne aa, der Breite der Fahne ss, dem Integral ber die *
 *     Konzentration quer zur Fahne mm und dem Anteil des Umwandlungspro-  *
 *     duktes uu abgespeichert.                                            *
 *                                                                         *
 *-------------------------------------------------------------------------*/

static int Tilted(/* Berechnung der Fahnen-Funktionen nach Tilted Plume */
int iq )          /* Index der Quelle                                   */
  { 
  float h0, xh, dh;
  float *cc, *aa, *ss, *mm, *uu, *cz, *ch;
  float ggp, ggm, gfp, gfm;
  float xi, x, dx, s0, s1, s2, p0, p1, p2, b, c, sp, uq;
  float fl, b10, b11, b12, b20, b21, b22;
  float gpi1, gpi2, gmi1, gmi2;
  float zpi1, zmi1, zpi2, zmi2, gzp, gzm;
  float hpi1, hmi1, hpi2, hmi2, ghp, ghm;
  float fpi0, fpi1, fpi2, fmi0, fmi1, fmi2;
  float aai0, aai1, aai2, qqi0, qqi1, qqi2, tti0, tti1, tti2;
  float dt, dma;
  float ii0i1, ii0i2, ii1i1, ii1i2, ii2i1, ii2i2;
  float jji1, jji2, sz, sy, ha, hg, hgi0;
  float hai2, hgi2, syi1, syi2, uqi0, uqi1, uqi2;
  int i, k, l, ni;
  if (VD < VS)  return -404;
  h0 = HQ(iq);
  xh = WF(iq);
  dh = UF(iq);
  aa = Aa[iq];
  cc = Cc[iq];
  cz = Cz[iq];
  ch = Ch[iq];
  ss = Sy[iq];
  mm = Mm[iq];
  uu = Uu[iq];

  if (!aa || !cc || !cz || !ch || !ss || !mm || !uu)  return -402;
  ni = NI(iq);
  dx = LI(iq);
  SigzMax = FZ*HM;
  ArgMax  = AM;
  DetMin  = DM;
  fpi0 = 1;
  fmi0 = 0;
  hgi0 = 0;
  aai0 = 1;
  aai1 = aai2 = aai0;
  qqi0 = 1;
  tti0 = 0;
  aa[0] = aai0;
  cc[0] = 0;
  cz[0] = 0;
  ch[0] = 0;
  ss[0] = 0;
  uu[0] = 0;
  sz   = 0;
  Wind( h0, hgi0, sz, aai0, &uqi0 );
  mm[0] = 1/uqi0;
  for (i=0; i<=ni-2; i+=2) {
    uqi2 = uqi0;
    for (l=0; l<2; l++) {
      fl = l + 1;
      x = (i + fl)*dx;
      if (x > xh)  ha = h0 + dh;
      else  ha = h0 + dh*pow(x/xh, 0.666667);
      if (ha > HM) ha = HM;
      uq = uqi0 + 0.5*fl*(uqi2 - uqi0);
      hg = hgi0 + fl*dx*VS*2.0/(uqi0 + uq);
      Sigma( x, ha-hg, &sz, TRUE, &sy );
      Teile(ha-hg, ha, hg, sz, HM, &ghp, &ghm, FALSE, NULL, NULL);
      Teile(ZP, ha, hg, sz, HM, &gzp, &gzm, FALSE, NULL, NULL);
      Teile( 0, ha, hg, sz, HM, &ggp, &ggm, TRUE, &gfp, &gfm );
      if (l == 0) {
        gpi1 = ggp;
        gmi1 = ggm;
        zpi1 = gzp;
        zmi1 = gzm;
        hpi1 = ghp;
        hmi1 = ghm;
        fpi1 = gfp;
        fmi1 = gfm;
        syi1 = sy; 
        uqi1 = uq; }
      else {
        gpi2 = ggp;
        gmi2 = ggm;
        zpi2 = gzp;
        zmi2 = gzm;
        hpi2 = ghp;
        hmi2 = ghm;
        fpi2 = gfp;
        fmi2 = gfm;
        hai2 = ha;
        hgi2 = hg;
        syi2 = sy; }
      s0 = 0;
      s1 = 0;
      s2 = 0;
      sp = 0;
      for (k=0; k<Ng[l]; k++) {
        xi = fl*Xg[l][k];
        x = dx*(i + xi);
        if (x > xh)  ha = h0 + dh;
        else  ha = h0 + dh*pow(x/xh, 0.666667);
        uq = uqi0 + 0.5*xi*(uqi2 - uqi0);
        hg = hgi0 + xi*VS*2./(uqi0 + uq);
        Sigma( x, ha-hg, &sz, FALSE, &sy );
        Teile( 0, ha, hg, sz, HM, &ggp, &ggm, FALSE, NULL, NULL);
        p0 = 1.0 - 0.5*xi*(3.0 - xi);
        p1 = xi*(2.0 - xi);
        p2 = 0.5*xi*(xi - 1.0);
        b = ggm*Wg[l][k]/uq;
        s0 = s0 + p0*b;
        s1 = s1 + p1*b;
        s2 = s2 + p2*b;
        sp = sp + ggp*Wg[l][k]/uq;
      }
      c = -fl*dx*VD;
      if (l == 0) {
        ii0i1 = c*s0;
        ii1i1 = c*s1;
        ii2i1 = c*s2;
        jji1 = c*sp; 
      }
      else {
        ii0i2 = c*s0;
        ii1i2 = c*s1;
        ii2i2 = c*s2;
        jji2 = c*sp; 
      }
    } 
    b10 = aai0*(ii0i1 + fmi0) + fpi0 - fpi1 + jji1;
    b11 = ii1i1 - fmi1;
    b12 = ii2i1;
    b20 = aai0*(ii0i2 + fmi0) + fpi0 - fpi2 + jji2;
    b21 = ii1i2;
    b22 = ii2i2 - fmi2;
    b = b11*b22 - b12*b21;
    if (ABS(b) > DetMin) {
      aai1 = (b20*b12 - b10*b22)/b;
      aai2 = (b21*b10 - b20*b11)/b; 
    }
    else {
      aai1 = aai0;
      aai2 = aai0; 
    }
    x = (i + 2)*dx;
    Sigma( x, hai2-hgi2, &sz, TRUE, &syi2 );
    Wind( hai2, hgi2, sz, aai2, &uqi2 );
/*
printf("WIND: x=%1.1f, ha=%1.1f, hg=%1.1f, sz=%1.1f, aa=%10.3e, uq=%1.2f\n",
  x, hai2, hgi2, sz, aai2, uqi2);  
*/
    uqi1 = 0.5*(uqi0 + uqi2);
    dt = dx*2/(uqi0+uqi1);
    tti1 = tti0 + dt;
    dma = AR*dt*qqi0;
    qqi1 = qqi0 - dma;
    dt = dx*2/(uqi1+uqi2);
    tti2 = tti1 + dt;
    dma = AR*dt*qqi1;
    qqi2 = qqi1 - dma;
    
    aa[i+1] = aai1;
    aa[i+2] = aai2;
    cc[i+1] = qqi1*(gpi1+aai1*gmi1)/uqi1;
    cc[i+2] = qqi2*(gpi2+aai2*gmi2)/uqi2;
    cz[i+1] = qqi1*(zpi1+aai1*zmi1)/uqi1;
    cz[i+2] = qqi2*(zpi2+aai2*zmi2)/uqi2;
    ch[i+1] = qqi1*(hpi1+aai1*hmi1)/uqi1;
    ch[i+2] = qqi2*(hpi2+aai2*hmi2)/uqi2;
    ss[i+1] = syi1;
    ss[i+2] = syi2;
    mm[i+1] = qqi1*(fpi1+aai1*fmi1)/uqi1;
    mm[i+2] = qqi2*(fpi2+aai2*fmi2)/uqi2;
    if (UR > 0) {
      c = tti1*UR;
      if (c < 1.e-4)  uu[i+1] = c;
      else  uu[i+1] = 1 - exp(-c);
      c = tti2*UR;
      if (c < 1.e-4)  uu[i+2] = c;
      else  uu[i+2] = 1 - exp(-c); 
    }
    else {
      uu[i+1] = 0;
      uu[i+2] = 0; 
    }
    hgi0 = hgi0 + 2.0*dx*VS/uqi1;
    uqi0 = uqi2;
    fpi0 = fpi2;
    fmi0 = fmi2;
    tti0 = tti2;
    qqi0 = qqi2;
    aai0 = aai2;
  }
  TstVertikal( x, hai2, hgi2, sz, aa[i] );    
  return 0; 
}

/*====================================================================== Boden
*/
static int Boden(   /* Berechnung der Konzentr. bei z=ZP */
int iq,             /* Index der Quelle                  */
float xa,           /* X-Koordinate des Aufpunktes       */
float ya,           /* Y-Koordinate des Aufpunktes       */
float si,           /* Sinus des Winkels Fahne/X-Achse   */
float co,           /* Kosinus des Winkels Fahne/X-Achse */
float *pwk,         /* Adresse Konzentrationswert (z=ZP) */
float *pwh,         /* Adresse Konzentrationswert (z=HA) */
float *pwn,         /* Adresse nasse Deposition          */
float *pwu )        /* Adresse umgewandelter Anteil      */
{ dP(Boden);
  float xp, yp, li, sy, ay, f, a1, a2;
  float *aa, *cz, *ch, *ss, *mm, *uu;
  int ni, ii, im;
  aa = Aa[iq];
  cz = Cz[iq];
  ch = Ch[iq];
  ss = Sy[iq];
  mm = Mm[iq];
  uu = Uu[iq];
  if (aa==NULL || cz==NULL || ss==NULL || mm==NULL || uu==NULL)  return -403;
  xa -= XQ(iq);
  ya -= YQ(iq);
  xp = co*xa + si*ya;
  yp = co*ya - si*xa;
  if (xp <= 1) {
    return 0; 
  }
  /*-22mar98-*/
  xp = sqrt(xp*xp + yp*yp);
  yp = xp*asin(yp/xp);
  /*---------*/
  ni = NI(iq);
  li = LI(iq);
  if (xp > ni*li)                                                    eX(1);
  ii = 1 + xp/li;
  if (ii > ni)  ii = ni;
  im = ii - 1; 
  a2 = (ii*li - xp)/li;
  a1 = 1 - a2;
  sy = a2*ss[im] + a1*ss[ii];
  ay = yp*yp/(2*sy*sy);
  if (ay > ArgMax) {
    return 0; 
  }
  f = EQ(iq)*F1*exp(-ay)/sy; 
  if (pwk) *pwk = f*(a2*cz[im] + a1*cz[ii]); 
  if (pwh) *pwh = f*(a2*ch[im] + a1*ch[ii]); 
  if (pwn) *pwn = AR*f*(a2*mm[im] + a1*mm[ii]);
  if (pwu) *pwu = a2*uu[im] + a1*uu[ii];
  return 1; 
eX_1:
  eMSG("*** Nicht genug Intervalle fr Aufpunkt xp=%f (ni=%d,li=%f)! (%s)\n", 
  xp, ni, li, ePGMs);  
} 
  
/*================================================================== DefRaster
*/
static int DefRaster( void )  /* Platz fr Rasterwerte anlegen */
  {
  dP(DefRaster);
  int nx, ny;
  Dauer = 0;
  nx = NX - 1;
  ny = NY - 1;  
  AryCreate(&ConDsc, sizeof(float), 2, 0, nx, 0, ny);    eG(1);
  AryCreate(&DryDsc, sizeof(float), 2, 0, nx, 0, ny);    eG(2);
  AryCreate(&WetDsc, sizeof(float), 2, 0, nx, 0, ny);    eG(3);
  AryCreate(&TrnDsc, sizeof(float), 2, 0, nx, 0, ny);    eG(4);
  AryCreate(&DtbDsc, sizeof(DTBREC), 2, 0, nx, 0, ny);   eG(5);
  AryCreate(&QtlDsc, sizeof(float), 2, 0, nx, 0, ny);    eG(6);
  return 0;
eX_1: eX_2: eX_3: eX_4: eX_5: eX_6:
  eMSG("*** Daten-Tabelle kann nicht angelegt werden! (%s)\n", ePGMs);
}

/*===================================================================== ClcCnc
*/
static int ClcCnc(     /* Berechnung der Konzentration      */
int iq,         /* Index der Quelle                  */
float xa,       /* X-Koordinate des Aufpunktes       */
float ya,       /* Y-Koordinate des Aufpunktes       */
float si,       /* Sinus des Winkels Fahne/X-Achse   */
float co,       /* Kosinus des Winkels Fahne/X-Achse */
float *pwk,     /* Adresse Konzentrationswert        */
float *pwi,     /* Intermittenz                      */
float *pwn,     /* Adresse nasse Deposition          */
float *pwu )    /* Adresse umgewandelter Anteil      */
  {
  float l1, l2, xq, yq, rq, siq, coq, xs, ys, d1, d2, d0;
  float a0, a1, a2, as, dx1, dy1, dx2, dy2, x0, y0, fq;
  float wk, wh, wn, wu;
  float wksum, wusum, wnsum, whsum;
  float *pwh;
  int i1, i2, rcm, n1, n2, rc;
  if (pwi)  pwh = &wh;
  else  pwh = NULL;
  rcm = 0;
  l1 = AQ(iq);
  l2 = BQ(iq);
  if (l1>0 || l2>0) {  /* Flchenquelle */
    xq = XQ(iq);
    yq = YQ(iq);
    rq = -RQ(iq);
    siq = sin( rq*PI/180 );
    coq = cos( rq*PI/180 );
    xs = (xa-xq)*coq - (ya-yq)*siq;
    ys = (ya-yq)*coq + (xa-xq)*siq;
    d1 = (xs > 0) ? xs - 0.5*l1 : -xs - 0.5*l1;
    d2 = (ys > 0) ? ys - 0.5*l2 : -ys - 0.5*l2;
    d0 = MAX(0,d1) + MAX(0,d2);
    if (d0 > 0) {
      a0 = (l1 + l2)* (1/d0 + 1/sqrt(L0*d0));
      if (a0 > A0)  a0 = A0; 
    }
    else  a0 = A0;
/*
printf("xa=%1.0f ya=%1.0f d1=%1.0f d2=%1.0f a0=%1.1f\n", xa, ya, d1, d2, a0);
*/
    if (l1 > l2) {
      as = sqrt(a0*l2/l1);
      a2 = MAX(1,as);
      a1 = a0/a2;
    }
    else {
      as = sqrt(a0*l1/l2);
      a1 = MAX(1,as);
      a2 = a0/a1; 
    }
    n1 = ceil(a1);
    n2 = ceil(a2);
/*
printf("n1=%d  n2=%d\n", n1, n2);
*/
    if (n1>1 || n2>1) {
      dx1 =  l1*coq/n1;
      dy1 = -l1*siq/n1; 
      dx2 =  l2*siq/n2;
      dy2 =  l2*coq/n2;
      x0  = xa - 0.5*((n1-1)*dx1 + (n2-1)*dx2);
      y0  = ya - 0.5*((n1-1)*dy1 + (n2-1)*dy2); 
    }
    else {
      x0 = xa;  dx1 = 0;  dx2 = 0;  
      y0 = ya;  dy1 = 0;  dy2 = 0; 
    }
    fq = 1.0/(n1*n2);
    wksum = 0;
    wusum = 0;
    whsum = 0;
    wnsum = 0;
    for (i1=0; i1<n1; i1++) {
      for (i2=0; i2<n2; i2++) {
        xs = x0 + i1*dx1 + i2*dx2;
        ys = y0 + i1*dy1 + i2*dy2;
/*
printf("%d,%d: xs=%1.0f ys=%1.0f\n", i1, i2, xs, ys);
*/
        rc = Boden( iq, xs, ys, si, co, &wk, pwh, &wn, &wu );
        if (rc < 0)  return rc;
        if (rc > 0) {
          wksum += wk;
          wnsum += wn;
          wusum += wu*wk;
          if (pwh)  whsum += *pwh;
          rcm = 1; 
        }
      } 
    }
    wksum *= fq;
    wnsum *= fq;
    wusum *= fq;
    if (pwh)  whsum *= fq;
  }
  else {   /* Punktquelle */ 
    rc = Boden( iq, xa, ya, si, co, &wk, pwh, &wn, &wu );
    if (rc < 0)  return rc;
    if (rc > 0) {
      wksum = wk;
      wnsum = wn;
      wusum = wu*wk;
      if (pwh)  whsum = *pwh;
      rcm = 1; 
    }
  }
  if (rcm) {
    if (pwk) *pwk = wksum;
    if (pwn) *pwn = wnsum;
    if (pwu) *pwu = wusum;
    if (pwi) {
      if (whsum <= wksum)  *pwi = 1;
      else  *pwi = wksum/whsum;
    }
  }
  return rcm;
}
    

/*================================================================== ClcRaster
*/ 
static int ClcRaster( void )   /* Werte auf dem Raster berechnen */
  {
  int i, j, ij, n, iq, iu, iw, rc, rcm;
  float dt, si, co, rw, dw, ru, du, r0, wk, wn, wu, xa, ya;
  float wksum, wtsum, wnsum, wusum;
  dw = 360/NW;
  du = dw/NU;
  for (iw=0; iw<NW; iw++) {
    rw = RE + iw*dw;
    dt = DW(iw)/NU;
    if (dt <= 0)  continue;
    r0 = rw - 0.5*(NU-1)*du;
    for (iu=0; iu<NU; iu++) {
      ru = r0 + iu*du;
      si = sin((270-ru)*PI/180);
      co = cos((270-ru)*PI/180);
      n = (NX>0 && NY>0) ? -1 : 0;
      for (i=0, j=0; ; ) {
          if (n<0 && j>=NY) {
            j = 0;
            i++;
            if (i >= NX)  n = 0;
          }
          if (n >= NP)  break;
          if (n >= 0) {
            xa = XP(n);
            ya = YP(n);
            if (xa==-999 && ya==-999) { n++;  continue; }
          }
          else {
            xa = X0 + i*DD;
            ya = Y0 + j*DD;
            ij = i*NY + j;
          }
          wksum = 0;
          wtsum = 0;
          wnsum = 0;
          wusum = 0;
          rcm = 0;
          for (iq=0; iq<NQ; iq++) {
            rc = ClcCnc(iq, xa, ya, si, co, &wk, NULL, &wn, &wu);
            if (rc < 0)  return rc;
            if (rc > 0) {
              wksum += wk;
              wnsum += wn;
              wusum += wu;
              wtsum += VD*wk;
              rcm = 1;
            }
          } /* iq */
          /*  Ausgabe von Einzelwerten fr Perzentilberechnung */
          if (EwtFile) {
            if (wksum > 0.0) {
              EwtRec.i = i;
              EwtRec.j = j;
              EwtRec.v = wksum;
              EwtRec.w = dt;
              if (1 != fwrite(&EwtRec, sizeof(EwtRec), 1, EwtFile)) return -444;
              Newt++;
            }
          }
          if (!rcm && n<0 && doQtl) 
            QtlRegister(0., dt, (DTBREC *)AryPtrX(&DtbDsc, i, j));
          if (rcm) {
            if (n < 0) {
              if (doCon)  CON(i,j) += dt*wksum;
              if (doDry)  DRY(i,j) += dt*wtsum;
              if (doWet)  WET(i,j) += dt*wnsum;
              if (doTrn)  TRN(i,j) += dt*wusum;
              if (doQtl) QtlRegister(wksum, dt, (DTBREC *)AryPtrX(&DtbDsc, i, j));
            }
            else {
              if (doCon)  Cona[n] += dt*wksum;
              if (doDry)  Drya[n] += dt*wtsum;
              if (doWet)  Weta[n] += dt*wnsum;
              if (doTrn)  Trna[n] += dt*wusum;
            }
          }
          if (n >= 0)  n++;
          else         j++;
        } /* for */
      Dauer += dt;
    }  /* iu */
  } /* iw */
  return 1; 
}    

  
//======================================================= WriteResult
static char *trmfixed( double x )
{
  static char s[80];
  int i;
  if (x>1.e6 || x<-1.e6)  sprintf(s, "%10.3le", x);
  else {
    sprintf(s, "%1.2lf", x);
    for (i=strlen(s)-1; i>0; i--) {
      if (s[i] == '0')  s[i] = 0;
      else {
        if (s[i] == '.')  s[i] = 0;
        break;
      }
    }
  }
  return s;
}

static int Normalize( ARYDSC *psrc, ARYDSC *pdst, float dt )
{ dP(Normalize);
  int i, j;
  void *pv;
  float *ps, *pd;
  if (dt > 0) {
    *pdst = *psrc;
    pdst->start = NULL;
    pv = ALLOC(pdst->ttlsz);  if (!pv)            eX(1);
    pdst->start = pv;
  }
  for (i=0; i<NX; i++)
    for (j=0; j<NY; j++) {
      ps = AryPtr(psrc, i, j);  if (!ps)        eX(2);
      pd = AryPtr(pdst, i, j);  if (!pd)        eX(3);
      if (dt > 0)  *pd = (*ps)/dt;
      else         *pd = (*ps > 0) ? *pd/(*ps) : 0;
    }
  return 0;
eX_1:
    eMSG("*** can't allocate work array! (%s)\n", ePGMs);
eX_2: eX_3:
    eMSG("*** index (%d,%d) out of range! (%s)\n", i, j, ePGMs);
}

static int WriteResult( char *arg )     // Rasterausgabe einleiten
{
  dP(WriteResult);
  int i, j, l;
  char tt[120], unit[32], s[80], fname[256], name[80];
  char *pc, *pn, *buf;
  float fa;
  ARYDSC dsc;
  TXTSTR usrhdr = { NULL, 0 };
  if (Dauer <= 0) {
    vMsg("Es sind keine Werte berechnet!\n");
    return 0;
  }
  if (NX<=0 || NY<=0) {
    vMsg("Es ist kein Gitter definiert!\n");
    return 0;
  }
  MsgLow(arg);
  pn = arg;
  pc = MsgTok(pn, &pn, " +\t\r\n", &l);
  while (pc) {
    *name = 0;
    strncat(name, pc, l);
    fa = 0;
    memset(&dsc, 0, sizeof(ARYDSC));
    if (!strcmp(name, "cnc")) {
      Normalize(&ConDsc, &dsc, Dauer);                               eG(10);
      sprintf(tt, "Konzentration");
      sprintf(unit, "%s/m", ME);
    }
    else if (!strcmp(name, "dry")) {
      Normalize(&DryDsc, &dsc, Dauer);                               eG(11);
      sprintf(tt, "Trockene Deposition");
      sprintf(unit, "%s/sm", ME);
    }
    else if (!strcmp(name, "wet")) {
      Normalize(&WetDsc, &dsc, Dauer);                               eG(12);
      sprintf(tt, "Nasse Deposition");
      sprintf(unit, "%s/sm", ME);
    }
    else if (!strcmp(name, "trn")) {
      Normalize(&TrnDsc, &dsc, Dauer);                               eG(13);
      sprintf(tt, "Umgewandelte Konzentration");
      sprintf(unit, "%s/m", ME);
    }
    else if (!strcmp(name, "qtl")) {
      if (0>QtlQuantiles(QT))                                        eX(14);
      Normalize(&QtlDsc, &dsc, 1);
      sprintf(s, "%1.2f", QT*100);
      sprintf(tt, "%s-Perzentil", s);
      sprintf(unit, "(%s/m)", ME);
      sprintf(s, "%05d", (int)(QT*10000+0.5));
      strcat(name, s);
    }
    else                                                             eX(15);
    if (RV > 0) {
      sprintf(s, " bezogen auf %1.2e %s", RV, unit);
      strcat(tt, s);
    }
    if (fa <= 0)  fa = FA;
    if (!strncmp(FI, "LOG", 3)) strcpy(fname, FI);
    else {
      strcpy(s, (*FI) ? FI : name);
      strcpy(fname, MakeFileName(s, ""));
    }
    TxtClr(&usrhdr);
    if (strlen(ID) > 0) {
      sprintf(s, "mark\t\"%s\"\n", TxtUnquote(ID)); TxtCat(&usrhdr, s);
    }
    /* else if (RV > 0) {
      sprintf(s, "mark\t\"%s\"\n", tt);             TxtCat(&usrhdr, s);
    }*/
    if (*TI) {
      sprintf(s, "title\t\"%s\"\n", TxtUnquote(TI)); TxtCat(&usrhdr, s);
    }
    if (*MO) {
      sprintf(s, "mode\t\"%s\"\n", TxtUnquote(MO)); TxtCat(&usrhdr, s);
    }
    if (*SE) {
      sprintf(s, "sequ\t\"%s\"\n", TxtUnquote(SE)); TxtCat(&usrhdr, s);
    }
    if (RV > 0) {
      sprintf(s, "refv\t%10.2e\n", RV); TxtCat(&usrhdr, s);
    } 
    if (*FO) {
      sprintf(s, "form\t\"%s\"\n", TxtUnquote(FO)); TxtCat(&usrhdr, s);
    }
    if (FA > 0) {
      sprintf(s, "fact\t%10.3e\n", FA); TxtCat(&usrhdr, s);
    }
    sprintf(s, "unit\t%s\n", unit);         TxtCat(&usrhdr, s);
    sprintf(s, "xmin\t%s\n", trmfixed(X0)); TxtCat(&usrhdr, s);
    sprintf(s, "ymin\t%s\n", trmfixed(Y0)); TxtCat(&usrhdr, s);
    sprintf(s, "delt\t%s\n", trmfixed(DD)); TxtCat(&usrhdr, s);
    sprintf(s, "vldf\t%s\n", "P");          TxtCat(&usrhdr, s);
    buf = TxtAlloc(4000);
    strcpy(buf, "ilbl");
    for (i=0; i<NX; i++) {
      strcpy(s, "\t");
      strcat(s, trmfixed(X0+i*DD));
      if (strlen(buf)+strlen(s) < 3998)  strcat(buf, s);
    }
    strcat(buf, "\n");            TxtCat(&usrhdr, buf);
    strcpy(buf, "jlbl");
    for (j=0; j<NY; j++) {
      strcpy(s, "\t");
      strcat(s, trmfixed(Y0+j*DD));
      if (strlen(buf)+strlen(s) < 3998)  strcat(buf, s);
    }
    strcat(buf, "\n");            TxtCat(&usrhdr, buf);
    FREE(buf);
    DmnWrite(fname, usrhdr.s, &dsc);                eG(1);
    AryFree(&dsc);
    TxtClr(&usrhdr);
    if (strncmp(FI, "LOG", 3))  *FI = 0;
    *ID = 0;
     pc = MsgTok(pn, &pn, " +\t\r\n", &l);
  }
  return 0;
eX_1:
  eMSG("*** can't write array %s! (%s)\n", name, ePGMs);
eX_10: eX_11: eX_12: eX_13:
  eMSG("*** can't normalize array %s! (%s)\n", name, ePGMs);
eX_14:
  eMSG("*** can't get quantiles! (%s)\n", ePGMs);
eX_15:
  vMsg("--- unknown array name %s! (%s)\n", name, ePGMs);
  return 0;
}

/*================================================================== PrnPoints
*/
static int PrnPoints( void )
  {
  int n, l;
  if (NP <= 0)  return 0;
  l = 3 - strlen(ME);
  fprintf(LogFile, "Werte an den Aufpunkten:\n");
  fprintf(LogFile, "  i:  Xp (m)  Yp (m)");
  if (doCon)  fprintf(LogFile, "%*s  Cm (%s/m3)", l, " ", ME); 
  if (doDry)  fprintf(LogFile, " %*sFd (%s/sm2)", l, " ", ME); 
  if (doWet)  fprintf(LogFile, " %*sFw (%s/sm2)", l, " ", ME); 
  if (doTrn)  fprintf(LogFile, "%*s  Tr (%s/m3)", l, " ", ME);
  fprintf(LogFile, "\n");
  for (n=0; n<NP; n++) {
    fprintf(LogFile, "%3d: %7.1f %7.1f", n+1, XP(n), YP(n));
    if (doCon)  fprintf(LogFile, " %12.3e", MMAX(Cona[n]/Dauer)); 
    if (doDry)  fprintf(LogFile, " %12.3e", MMAX(Drya[n]/Dauer)); 
    if (doWet)  fprintf(LogFile, " %12.3e", MMAX(Weta[n]/Dauer)); 
    if (doTrn)  fprintf(LogFile, " %12.3e", MMAX(Trna[n]/Dauer)); 
    fprintf(LogFile, "\n");
  }
  fprintf(LogFile, "\n");
  return 0;
}

/*=================================================================== ClearAll
*/
static int ClearAll( void )
  {
  int n;
  n = NX*NY*sizeof(float);
  if (ConDsc.start)  memset(ConDsc.start, 0, n);
  if (DryDsc.start)  memset(DryDsc.start, 0, n);
  if (WetDsc.start)  memset(WetDsc.start, 0, n);
  if (TrnDsc.start)  memset(TrnDsc.start, 0, n);
  if (QtlDsc.start)  memset(QtlDsc.start, 0, n);
  if (DtbDsc.start)  QtlClear();
  n = NP*sizeof(float);
  if (Cona)  memset(Cona, 0, n);
  if (Drya)  memset(Drya, 0, n);
  if (Weta)  memset(Weta, 0, n);
  if (Trna)  memset(Trna, 0, n);
  Dauer = 0;
  return 0;
}
  
//============================================================ MyExit
static void MyExit( void )
{
  time_t tm;
  time(&tm);
  MsgVerbose = 1;
  MsgQuiet--;
    if (MsgCode < 0)  vMsg("*** error exit");
    else              vMsg("FINISHED");
  vMsg(" at %s", ctime(&tm));
}

   
/*======================================================================= main
*/
int main(           /* Hauptprogramm, Steuerung der Verfahrensschritte */
int argc, 
char *argv[] )      /* Name des Eingabefiles, Optionen                 */
  {
  dP(Main);
  int rc, i, iq, done, l;
  char *pk, *s, fname[MAXFNLEN], tms[80];
  char cmd[200], arg[PRM_BUFLEN], opt[32];
  char buf[200];
  time_t tm;
  static char alles[] = "*c*d*g*l*p*q*l*w*s*t*e";
  atexit(MyExit);
   if (argc < 2) {
    printf("Information zur Art des Aufrufes: IBJplume -h\n");
    exit(0);
  }
  MsgVerbose=0;
  for (i=1; i<argc; i++) {
    s = argv[i];
    if (strchr(SYMOPT, *s)) 
      switch (s[1]) {
        case 'e': PrmEcho = 1;
                  sscanf(s+2, "%x", &PrmEcho);
                  break;
        case 'h': Help();
        case 'i': strcpy(CmdName, s+2);
                  break;
        case 'l': strcpy(LogName, s+2);
                  break;
        case 'q':  MsgQuiet = 1;
                   sscanf(s+2, "%x", &MsgQuiet);
                   break;
        case 'v':  Verbose = 1;
                   break;
        default:   printf("Unbekannte Option %s\n", s);
      }
    else {
      strcpy(WorkPath, s);
      for (pk=WorkPath; (*pk); pk++)  if (*pk == ALTPSEP)  *pk = PATHSEP;
      l = strlen(WorkPath);
      if ((WorkPath[l-1]!=':') && (WorkPath[l-1]!=PATHSEP)) {
        WorkPath[l] = PATHSEP;
        WorkPath[l+1] = 0;
        }
      }
    }
  if (*LogName != '-') {
    strcpy(fname, MakeFileName(LogName, ".log"));                    eG(26);
    MsgFile = fopen(fname, "w");  if (!MsgFile)                      eX(27);
    LogFile = MsgFile;
    }
  time(&tm);
  strcpy(tms, ctime(&tm));
  l = strlen(tms)-1;
  tms[l] = 0;
  MsgVerbose = 1;
  vMsg("\nIBJplume: Gaussian plume model, version %s\n", VERSION);
  vMsg("Copyright (C) Janicke Consulting, Dunum, Germany, 2001\n");
  vMsg("IBJplume comes with ABSOLUTELY NO WARRANTY\n");
  vMsg("IBJplume is free software under the GNU PUBLIC LICENCE\n");
  fprintf(MsgFile, "Compiled at %s  %s\n", __DATE__, __TIME__);
  vMsg("STARTING at %s", ctime(&tm));
  vMsg("Path: %s, Input file: %s\n\n", WorkPath, CmdName);
  //rc = Init(0);
  if (rc < 0)  done = TRUE;
  else         done = FALSE;
  time(&tm);
  *cmd = 0;
  *arg = 0;  
  strcpy(fname, MakeFileName(CmdName, ".txt"));
  while (PrmRead(fname, cmd, arg)) {                                 eG(2);
    *fname = 0;
    if (!Initialized && tolower(cmd[1])!='d')                        eX(16);
    switch (tolower(cmd[1])) {
      case '*': if (*arg) fprintf(MsgFile, "%s\n", arg);
                break;
      case 'a': 
                break;
      case 'c': ClearAll(); 
                break;
      case 'd': strcpy(Title, arg);
                Init(0);                                  
                break;
      case 'g': DefRaster();                             
                break;
      case 'p': Stat();
                break;
      case 'q': ClrPlume(0);   
                break;
      case 's': WriteResult(arg);                                    eG(11);
                break;
      case 't': if (NULL == strcmp(arg, "*"))
                  PrmPrint(alles);
                else
                  for (i=0; i<strlen(arg); i++) {
                    strcpy(buf, "*");
                    if (arg[i] != '*')  strncat(buf, arg+i, 1);
                    PrmPrint(buf);
                  }
                break;
      case 'l': if (!*arg) break;
                if (Dauer <= 0) Stat();
                if (strstr(arg, "plm")) {
                  for (iq=0; iq<NQ; iq++) {
                    rc = PrnPlume(iq);
                    if (rc < 0)                                      eX(14);
                  }
                }
                if (strstr(arg, "pnt"))  
                  PrnPoints();
                break;
      case 'w': if (!*arg)                                           eX(17);
                if (PrmSet("hm"))  HM = 0;
                strcpy(opt, "");
                if (RH > 0) strcpy(opt, "g");
                if (MODUS(TALUFT)) strcpy(opt, "T");
                ClearAll();
                SttWork(WorkPath, opt, arg, NULL, 2, 0, Stat);        eG(18);
                break;
      case 'e': done = TRUE;
                break;
      default:                                                       eX(15);
    }
    if (done)  break;
  }  // while
  PrmRead(NULL, NULL, NULL);                                         eG(2);
  return 0;
eX_2:
  eMSG("*** Fehler beim Lesen des Eingabe-Files! (%s)\n", ePGMs);
eX_11:
  eMSG("*** Fehler beim Ausschreiben der Ergebnisse! (%s)\n", ePGMs);
eX_14:
  eMSG("*** Fehler beim Ausschreiben der Fahnenparameter! (%s)\n", ePGMs);
eX_15:
  eMSG("*** Unbekannter Abschnitt %s! (%s)\n", cmd, ePGMs);
eX_16:
  eMSG("*** Fehlerhafte Initialisierung! (%s)\n", ePGMs);
eX_17:
  eMSG("*** Name der Wetterstatistik fehlt! (%s)\n", ePGMs);
eX_18:
  eMSG("*** Fehler beim Abarbeiten der Wetterstatistik! (%s)\n", ePGMs);
eX_26:
  eMSG("*** Name fr Log-File kann nicht gebildet werden! (%s)\n", ePGMs);
eX_27:
  eMSG("*** Log-File %s nicht beschreibbar! (%s)\n", fname, ePGMs);
}
  
/*==========================================================================*/
