//======================================================== IBJqtl.c
//
// Calculation of Quantiles
// ========================
//
// Copyright (C) Janicke Consulting, Dunum, Germany, 2001
// email: info@janicke.de
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// last change: 2001-08-01 uj
//
//==================================================================

char *IBJqtlVersion = "1.2";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "IBJary.h"
#include "IBJmsg.h"
#include "IBJqtl.h"

ARYDSC DtbDsc;
ARYDSC QtlDsc;

//----------------------------------------------------------- QtlValue
float QtlQuantile(float q, DTBREC *p) {
  float sum, qsum, psum;
  int i, k, nf;
  float w, wmin;
  double a, arg;
  nf = DTB_NUMVAL;  
  if (q >= 1.0) {
    return p->frq[nf]; 
  }
  sum = p->frqnull + p->frqsub;
  for (i=0; i<nf; i++)  sum += p->frq[i];
  qsum = q*sum;
  if (qsum <= p->frqnull)  return 0.0;
  if (qsum < p->frqnull + p->frqsub) {
    wmin = p->valmax*pow(10.0,-nf/DTB_DIVISION);
    return -wmin; 
  }
  qsum = sum - qsum;
  psum = 0;
  for (k=nf-1; k>=0; k--) {
    psum += p->frq[k];
    if (psum >= qsum)  break;
  }
  if (k < 0) {
    wmin = p->valmax*pow(10.0,-nf/DTB_DIVISION);
    return wmin; 
  }
  a = (psum - qsum)/p->frq[k];
  if (k == nf-1)
    arg =(a - 1)/DTB_DIVISION + a*log10(p->frq[nf]/p->valmax);
  else  
    arg = (k + a - nf)/DTB_DIVISION;
  w = p->valmax*pow(10.0, arg);
  return w; 
}


//---------------------------------------------------------- QtlSetValues
int QtlQuantiles(float q) {
  int i, j, n;
  if (!QtlDsc.start) return -1;
  if (!DtbDsc.start) return -2;
  if (QtlDsc.numdm != DtbDsc.numdm) return -3;
  n = DtbDsc.numdm;
  if (n != 2) return -3;
  if (DtbDsc.bound[0].low != QtlDsc.bound[0].low) return -4;
  if (DtbDsc.bound[0].hgh != QtlDsc.bound[0].hgh) return -4;
  if (DtbDsc.bound[1].low != QtlDsc.bound[1].low) return -4;
  if (DtbDsc.bound[1].hgh != QtlDsc.bound[1].hgh) return -4;
  for (i=DtbDsc.bound[0].low; i<=DtbDsc.bound[0].hgh; i++) {
    for (j=DtbDsc.bound[1].low; j<=DtbDsc.bound[1].hgh; j++) {
      *(float*)AryPtrX(&QtlDsc, i, j) = 
      QtlQuantile(q, (DTBREC *)AryPtrX(&DtbDsc, i, j));
    }
  }
  return 0;
}
  

//----------------------------------------------------------- QtlRegister
int QtlRegister(float a, float g, DTBREC *p) {
  int i, n, m;
  float g0;
  if (a > p->realmax) {   
    p->realmax = a;
  }
  if (a <= 0.0)  {
    p->frqnull += g;
    return 0; 
  }
  if (a >= DTB_RANGE*p->valmax) {
    g0 = 0.0;
    for (i=0; i<DTB_NUMVAL; i++) {
      g0 += p->frq[i];
      p->frq[i] = 0.0; 
    }
    p->frqsub += g0;
    n = floor(1 + DTB_DIVISION*log10(a));
    p->valmax = pow(10.0, ((float)n)/DTB_DIVISION);
    p->frq[DTB_NUMVAL-1] = g;
    return 0;
  }
  n = floor(DTB_NUMVAL + DTB_DIVISION*log10(a/p->valmax));
  if (n < 0)  {
    p->frqsub += g;
    return 0;
  }
  if (n < DTB_NUMVAL) {
    p->frq[n] += g;
    return 0; 
  } 
  m = n - DTB_NUMVAL + 1;
  g0 = 0.0;
  for (i=0; i<m; i++)  g0 += p->frq[i];
  p->frqsub += g0;
  for ( ; i<DTB_NUMVAL; i++) {
    p->frq[i-m] = p->frq[i];
    p->frq[i] = 0.0; 
  }
  p->valmax *= pow(10.0, m/DTB_DIVISION);
  p->frq[DTB_NUMVAL-1] = g;
  return 0; 
}

//-------------------------------------------------------------- QtlClear
int QtlClear( void ) {
  int n, i, j, k;
  DTBREC pa;
  n = DtbDsc.numdm;
  if (n != 2) return -1;
  for (i=DtbDsc.bound[0].low; i<=DtbDsc.bound[0].hgh; i++) {
    for (j=DtbDsc.bound[1].low; j<=DtbDsc.bound[1].hgh; j++) {
      pa = DTB(i, j);
      pa.valmax = 0.;
      pa.frqnull = 0.;
      pa.frqsub = 0.;
      pa.valmax = 0.;
      pa.realmax = 0.;
      pa.realdev = 0.;
      for (k=0; k<DTB_NUMVAL; k++) pa.frq[k] = 0.;
    }
  }
  return 0;
}


#ifdef  MAIN //#############################################

int main( int argc, char *argv() )
  {
  return 0;
  }
#endif  //##################################################

/*==============================================================
//
// history:
//
//=============================================================*/
