//======================================================== IBJprm.c
//
// New Input Routine for IBJ-programs
// ==================================
//
// Copyright (C) Janicke Consulting, Dunum, Germany, 1998-2001
// email: info@janicke.de
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// history:
//
// 98-07-31   lj
// 98-08-31   lj  increment for reference value added
//
//==================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>

char *PrmVersion = "1.1.e";

#define  IBJPRM_SOURCE

#include "IBJprm.h"
#include "IBJtbl.h"
#include "IBJmsg.h"

#define  ISARRAY(pt)    (pt->type & PRM_ISARRAY)
#define  ISSTRING(pt)   ((pt->type & PRM_ISANY) == PRM_ISSTRING)
#define  ISINT(pt)      ((pt->type & PRM_ISANY) == PRM_ISINT)
#define  ISCOMMAND(pt)  (pt->name[0] == '*')

typedef struct prmfrec *PRMFPTR;
typedef struct prmfrec {
  PRMFPTR last;
  char fns[256];
  FILE *f;
  int line_count;
  int loop_start;
  int loop_count;
  int loop_lncnt;
} PRMFREC;

int PrmBufSize = 1024;
int PrmValid = 2;         // number of significant characters
char PrmLabel[200];
int PrmEcho;

static char *eMODn = "IBJprm";
static int InitLevel;

static void trim( char * s )
  {
  char *p, *q;
  int instring = -1;
  for (p=s, q=s; *q<=' ' && *q>0; q++);
  if (*q == '"')  instring = 0;
  for ( ; (*q); q++) {
    if (*q == '"') {
      if (instring) {
        if (q[1] == '"')  q++;
        else {
          instring = 0;
          continue;
        }
      }
      else if (instring == 0) {
        instring = 1;
        continue;
      }
    }
    *p++ = *q;
  }
  *p = 0;
  for (p--; p>=s; p--)
      if (*p <= ' ')  *p = 0;
      else  break;
  }

//============================================================== PrmInit
int PrmInit(    // Initialize the data structure PrmTab
  int level )   // 1: set scalars and strings only
{         // 2: allocate and initialize vectors
  dP(PrmInit);
  int i, n, l, dn;
  void *p;
  char *pc, **ppc, *m;
  PRMTABREC *pt, *qt;
  for (pt=(PRMTABREC*)&PrmTab; (pt->input_format) ; pt++) {
    if (ISCOMMAND(pt))  continue;
    if (pt->type & PRM_DATSET)  continue;
    m = pt->name;
    if (ISARRAY(pt)) {
      if (level < 2)  continue;
      p = pt->u.p;
      if (p)  free(p);
      l = pt->size;
      if (ISSTRING(pt))  l += sizeof(char*)+1;
      n = (*pt->reference) ? -1 : 0;
      sscanf(pt->reference, "%d", &n);  if (!n)     eX(1);
      if (n < 0) {
        for (qt=(PRMTABREC*)&PrmTab; qt<pt ; qt++)
          if (!strncmp(pt->reference, qt->name, PrmValid))  break;
        if (qt >= pt)                               eX(2);
        if (ISARRAY(qt) || (!ISINT(qt)))            eX(3);
        dn = 0;                                     //-98-08-31
        sscanf(pt->reference+PrmValid, "%d", &dn);  //-98-08-31
        n = qt->u.i + dn;                           //-98-08-31
        if (n < 0)                                  eX(4);
      }
      if (n <= 0) {                                 //-99-01-26
        pt->u.p = NULL;
        pt->number = 0;
        continue;
      }
      p = calloc(n, l);  if (!p)            eX(5);
      pt->u.p = p;
      pt->number = n;
      if (ISSTRING(pt)) {
        ppc = p;
        pc = (char*)p + n*sizeof(char*);
        for (i=0; i<n; i++)  ppc[i] = pc + i*(pt->size+1);
        //  sscanf(pt->default_value, pt->input_format, ppc[0]);
        strcpy(ppc[0], pt->default_value);
        for (i=1; i<n; i++)  strcpy(ppc[i], ppc[0]);
      }
      else {
        pc = p;
        sscanf(pt->default_value, pt->input_format, p);
        for (i=1; i<n; i++)  memcpy(pc+i*l, pc, l);
      }
    }
    else {  // no array
      if (InitLevel > 0)  continue;
      if (ISSTRING(pt)) {
        p = pt->u.p;
        if (p)  free(p);
        p = calloc(1, pt->size+1);  if (!p)     eX(6);
        pt->u.p = p;
        strcpy(p, pt->default_value);
      }
      else {
        p = &(pt->u);
        sscanf(pt->default_value, pt->input_format, p);
      }
    }
  }
  InitLevel = level;
  return 0;
eX_1:
  eMSG("*** array %s: invalid number of elements! (%s)\n", m, ePGMs);
eX_2:
  eMSG("*** array %s: reference %s not found! (%s)\n",
    m, pt->reference, ePGMs);
eX_3:
  eMSG("*** array %s: reference %s must be an integer! (%s)\n",
    m, pt->reference, ePGMs);
eX_4:
  eMSG("*** array %s: reference %s has invalid value (=%d)! (%s)\n",
    m, pt->reference, n, ePGMs);
eX_5:
  eMSG("*** array %s: can't allocate (%d times %d bytes)! (%s)\n",
    m, n, l, ePGMs);
eX_6:
  eMSG("*** parameter %s: can't allocate string space (%d bytes)! (%s)\n",
    m, pt->size, ePGMs);
}

//============================================================= PrmReset
int PrmReset(     // Reset the data structure PrmTab
void )        // free or clear all parameters
{
  PRMTABREC *pt;
  for (pt=(PRMTABREC*)&PrmTab; (pt->input_format); pt++) {
    if (ISCOMMAND(pt))  continue;
    if ((ISARRAY(pt) || ISSTRING(pt)) && (pt->u.p))  free(pt->u.p);
    memset(&(pt->u), 0, sizeof(pt->u));
    pt->number = 0;
    pt->type &= ~PRM_DATSET;
  }
  InitLevel = 0;
  return 0;
}

//=============================================================== PrmFind
PRMTABREC *PrmFind( // Find a parameter in PrmTab
  char *name,   // name of the parameter
  char *cmd )   // name of the section
{
  PRMTABREC *pt;
  int section_found;
  section_found = (*cmd == '?');
  for (pt=(PRMTABREC*)&PrmTab; (pt->input_format); pt++) {
    if (*pt->name == '*') {
      if (section_found && (*cmd != '?'))  return NULL;
      else if (strstr(pt->name, cmd))  section_found = 1;
    }
    else if (section_found && !strncmp(pt->name, name, PrmValid))  return pt;
  }
  return NULL;
}

//=============================================================== PrmSet
int PrmSet(     // Returns 1, if the parameter ist set by the user
  char *name )  // name of the parameter
{
  PRMTABREC *pt;
  pt = PrmFind(name, "?");
  if (!pt)  return -1;
  if (pt->type & PRM_DATSET)  return 1;
  return 0;
}

//============================================================ PrmPrnRec
int PrmPrnRec(    // Print a record value into a string
char *b,          // string buffer
int maxlen,       // length of buffer
PRMTABREC *pt )   // address of selected record
{
  char z[400];
  int i, l, len, n;
  void *p;
  len = strlen(b);
  sprintf(z, "%-4s", pt->name);
  l = strlen(z);  if (l+len > maxlen)  return -1;
  strcat(b, z);
  if (ISARRAY(pt)) { p = pt->u.p;  n = pt->number; }
  else             { p = &(pt->u);  n = 1; }
  for (i=0; i<n; i++) {
    switch (pt->type & PRM_ISANY) {
      case PRM_ISINT:
        sprintf(z, pt->output_format, ((int*)p)[i]);
        break;
      case PRM_ISFLOAT:
        sprintf(z, pt->output_format, ((float*)p)[i]);
        break;
      case PRM_ISDOUBLE:
        sprintf(z, pt->output_format, ((double*)p)[i]);
        break;
      case PRM_ISSTRING:
        sprintf(z, pt->output_format, ((char**)p)[i]);
        break;
      default:
        sprintf(z, " %12s", "???");
      }
    l = strlen(z);  if (l+len > maxlen)  return -1;
    strcat(b, z);
    len += l;
  }
  strcat(b, "\n");
  return n;
}

//=========================================================== PrmPrnList
int PrmPrnList( // Print a list of values into a string
char *b,        // string buffer
int maxlen,     // buffer length
char *cmd,      // name of selected section
... )           // parameter names or new section (NULL ends)
{
  dP(PrmPrnList);
  va_list arg;
  char *name, sct[80];
  int n;
  PRMTABREC *pt;
  strcpy(sct, cmd);
  va_start(arg, cmd);
  for (n=0; ; n++) {
    name = va_arg(arg, char*);
    if (name == NULL)  break;
    if (*name == '*') {
      strcpy(sct, name);
      continue;
    }
    pt = PrmFind(name, sct);  if (!pt)      eX(1);
    if (0 > PrmPrnRec(b, maxlen, pt))     eX(2);
  }
  va_end(arg);
  return n;
eX_1:
  eMSG("*** parameter \"%s\" not found in setion \"%s\"! (%s)\n",
    name, sct, ePGMs);
eX_2:
  eMSG("*** buffer overflow! (%s)\n", ePGMs);
}

//============================================================= PrmPrint
int PrmPrint(   // Print all data in PrmTab
char *sec )     // selected section
{
  dP(PrmPrint);
  PRMTABREC *pt;
  int active, n;
  char buf[200], *pb;
  pb = calloc(1, 32004);
  active = (sec == NULL);
  if (InitLevel < 2)
    vMsg("NAME        VALUE  DIM DESCRIPTION\n");
  else
    vMsg("NAME     VALUE(S)\n");
  for (pt=(PRMTABREC*)&PrmTab; (pt->input_format); pt++) {
    if ((sec) && (*pt->name=='*'))
      active =  (NULL != strstr(sec, pt->name));
    if (!active)  continue;
    if (*pt->name == '*')
      vMsg("%-4s %s\n", pt->name, pt->description);
    else if (InitLevel < 1)
      vMsg("%-4s %12s %4s %s\n", pt->name,
        pt->default_value, pt->reference, pt->description);
    else if (InitLevel == 1) {
      if (ISARRAY(pt))
        vMsg("%-4s %12s %4s %s\n", pt->name,
          pt->default_value, pt->reference, pt->description);
      else {
        switch (pt->type & PRM_ISANY) {
          case PRM_ISINT:
            sprintf(buf, pt->output_format, pt->u.i);
            break;
          case PRM_ISFLOAT:
            sprintf(buf, pt->output_format, pt->u.f);
            break;
          case PRM_ISDOUBLE:
            sprintf(buf, pt->output_format, pt->u.d);
            break;
          case PRM_ISSTRING:
            sprintf(buf, pt->output_format, pt->u.p);
            break;
          default:
            strcpy(buf, "???");
        }
        vMsg("%-4s %12s %4s %s\n", pt->name,
          buf, pt->reference, pt->description);
      }
    }
    else {
      *pb = 0;
      n = PrmPrnRec(pb, 32000, pt);
      vMsg("%s", pb);
      if (n < 0)                  eX(2);
    }
  }
  free(pb);
  return 0;
eX_2:
  eMSG("*** buffer overflow at %s! (%s)\n", pt->name, ePGMs);
}

//=============================================================== PrmRead
int PrmRead(    // Read command file
  char *fn,     // file name (complete with path and extension)
  char *cmd,    // returned: first command found
  char *arg )   // returned: argument part
{
  dP(PrmRead);
  static PRMFREC *pfrec, *pfold;
  static FILE *f;
  static char *pb, *fns;
  static int loop_count, loop_start, loop_lncnt, line_count;
  static int buffer_size;
  char *pc, *pn, *pv, tokbuf[PRM_BUFLEN], label[80];
  int n, l, le, i, ni, reading, file_position;
  void *p;
  PRMTABREC *pt;
  if (!fn) {
    while (pfrec) {
      f = pfrec->f;
      if (f)  fclose(f);
      pfold = pfrec->last;
      FREE(pfrec);
      pfrec = pfold;
    }
    f = NULL;
    return 0;
  }
  if (!pb) {
    pb = ALLOC(PrmBufSize);  if (!pb)         eX(4);
    buffer_size = PrmBufSize;
  }
  else if (buffer_size < PrmBufSize) {
    char *pnew;
    pnew = ALLOC(PrmBufSize);  if (!pnew)       eX(4);
    memset(pnew, 0, PrmBufSize);
    memcpy(pnew, pb, buffer_size);
    FREE(pb);
    pb = pnew;
    buffer_size = PrmBufSize;
  }

  if (*fn) {
    if (pfrec) {
      pfrec->f = f;
      pfrec->line_count = line_count;
      pfrec->loop_count = loop_count;
      pfrec->loop_start = loop_start;
      pfrec->loop_lncnt = loop_lncnt;
    }
    pfold = pfrec;
    pfrec = ALLOC(sizeof(PRMFREC));  if (!pfrec)    eX(25);
    pfrec->last = pfold;
    fns = pfrec->fns;
    strcpy(fns, fn);
    line_count = 0;
    loop_start = 0;
    loop_count = 0;
    if (!strcmp(fns, "CON"))  f = stdin;
    else {
      f = fopen(fns, "rb");  if (!f)          eX(1);
    }
  }
  if (InitLevel < 1) {
    PrmInit(1);                     eG(0);
  }
  if (!f)                         eX(20);
  n = 0;
  *cmd = 0;
  *arg = 0;
  for (reading=2; (reading); )  {
    *pb = 0;
    file_position = ftell(f);
    if (!fgets(pb, PrmBufSize, f)) {
      if (f != stdin)  fclose(f);
      f = NULL;
      pfold = pfrec->last;
      FREE(pfrec);
      pfrec = pfold;
      if (pfrec) {
        f = pfrec->f;
        fns = pfrec->fns;
        line_count = pfrec->line_count;
        loop_start = pfrec->loop_start;
        loop_count = pfrec->loop_count;
        loop_lncnt = pfrec->loop_lncnt;
      }
      if (!*cmd && (f))  continue;
      reading = 0;
      break;
    }
    line_count++;
    if (strlen(pb) >= PrmBufSize-1)           eX(2);
    for (i=strlen(pb)-1; i>=0; i--)
      if (pb[i] <= ' ')  pb[i] = 0;
      else break;
    if (line_count==1 && !strncmp(pb,"-:",2)) {
      l = PrmBufSize;
      sscanf(pb+2, "%s %d", label, &l);
      if (strncmp(label,PrmLabel,strlen(PrmLabel))) eX(10);
      if (l > PrmBufSize) {
        PrmBufSize = l;
        FREE(pb);
        pb = ALLOC(PrmBufSize);  if (!pb)     eX(24);
      }
      continue;
    }
    if (*pb == '*') {
      if (*cmd) {
        reading = 0;
        if (pb[1] > ' ') {
          if (fseek(f, file_position, SEEK_SET))  eX(21);
          line_count--;
          break;
        }
      }
    }
    if (PrmEcho)  vMsg("%03d> %s\n", line_count, pb);
    if (!reading)  break;
    if (*pb == '-')  continue;
    if (*pb <= ' ')  continue;
    // pc = strchr(pb, '\'');
    // if (pc)  *pc = 0;
    for (l=0, pc=pb; (*pc); pc++)
      if (*pc == '"')  l = 1-l;
      else if ((!l) && (*pc=='\''))  break;
    *pc = 0;
    if (*pb == '*') {
      if (pb[1] <= ' ')  continue;
      if (strlen(pb) >= PRM_BUFLEN-1)         eX(6);
      pc = strpbrk(pb, " \t");
      if (pc) {
        *pc++ = 0;
        trim(pc);
        strcpy(arg, pc);
      }
      else  *arg = 0;
      trim(pb);
      MsgLow(pb);
      strcpy(cmd, pb);
      if (!strcmp(cmd, "*loop")) {
        loop_start = ftell(f);
        loop_lncnt = line_count;
        sscanf(arg, "%d", &loop_count);
        *cmd = 0;
        continue;
      }
      else if (!strcmp(cmd, "*next")) {
        if (loop_count > 0) {
          if (fseek(f, loop_start, SEEK_SET))     eX(22);
          loop_count--;
          line_count = loop_lncnt;
        }
        *cmd = 0;
        continue;
      }
      cmd[PrmValid] = 0;
      n++;
      continue;
    }

    if (!*cmd)  continue;
    n++;
    if (strlen(pb) >= PrmBufSize-1)             eX(5);
    pv = MsgTok(pb, &pn, " ;\t\r\n", &l);  if (!pv)     eX(7);
    if (l<1 || l>=PRM_BUFLEN)               eX(8);
    *tokbuf = 0;  strncat(tokbuf, pv, l);
    MsgLow(tokbuf);
    pt = PrmFind(tokbuf, cmd);  if (!pt)          eX(9);
    if (ISARRAY(pt)) {
      if (InitLevel < 2)                  eX(16);
      pc = pt->u.p;
      ni = pt->number; }
    else {
      pc = (char*)&(pt->u);
      ni = 1; }
    le = pt->size;
    for (i=0; i<ni; i++) {
      pv = MsgTok(pn, &pn, " ;\t\r\n", &l);  if (!pv)   eX(11);
      if (l<0 || l>=PRM_BUFLEN-1)             eX(12);
      *tokbuf = 0;  strncat(tokbuf, pv, l);
      if (ISSTRING(pt)) {
        if (l > le)                   eX(13);
        p = (ISARRAY(pt)) ? ((char**)pt->u.p)[i] : pt->u.p;
        strcpy(p, tokbuf);
      }
      else {
        p = pc + i*le;
        if (1 != sscanf(tokbuf, pt->input_format, p)) eX(14);
      }
    }
    pt->type |= PRM_DATSET;
  }

  if (InitLevel < 2) {
    PrmInit(2);                   eG(17);
  }

  return n;
eX_0: eX_16: eX_17: eX_20: eX_25:
  eMSG("*** internal error! (%s)\n", ePGMs);
eX_1:
  eMSG("*** can't open file %s! (%s)\n", fns, ePGMs);
eX_2: eX_5: eX_6: eX_8:
  eMSG("*** buffer overflow in line %d of %s! (%s)\n", line_count, fns, ePGMs);
eX_4: eX_24:
  eMSG("*** can't allocate data buffer (%d bytes)! (%s)\n", PrmBufSize, ePGMs);
eX_7:
  eMSG("*** can't analyse line %d in %s! (%s)\n", line_count, fns, ePGMs);
eX_9:
  eMSG("*** name \"%s\" unknown in section \"%s\"! (%s)\n", tokbuf, cmd, ePGMs);
eX_10:
  eMSG("*** invalid command file label (\"%s\" required)! (%s)\n",
    PrmLabel, ePGMs);
eX_11:
  eMSG("*** insufficient data in line %d of %s! (%s)\n", line_count, fns, ePGMs);
eX_12:
  eMSG("*** maximum token length exceeded in line %d of %s! (%s)\n",
    line_count, fns, ePGMs);
eX_13:
  eMSG("*** maximum string length exceeded in line %d of %s! (%s)\n",
    line_count, fns, ePGMs);
eX_14:
  eMSG("*** invalid data in line %d of %s! (%s)\n", line_count, fns, ePGMs);
eX_21: eX_22:
  eMSG("*** can't reposition input file %s! (%s)\n", fns, ePGMs);
}

#ifdef  MAIN  //============================================================

int main( int argc, char *argv[] ) {
  dP(PrmMain);
  char fn[256], cmd[204], arg[204];
  printf("TEST von IBJparam\n");
  if (argc > 1)  strcpy(fn, argv[1]);
  else strcpy(fn, ".");
  strcat(fn, "\\param.txt");
  strcpy(PrmLabel, "IBJparam");
  printf("\n before initialization:\n");
  PrmPrint(stdout, "");
  while (PrmRead(fn, cmd, arg)) {           eG(2);
    *fn = 0;
    printf("\nREAD:%s|%s|\n", cmd, arg);
  }
  PrmRead(NULL, cmd, arg);
  printf("\nafter input processing:\n");
  PrmPrint(stdout, "");
  *arg = 0;
  PrmPrnList(arg, 200, "*d", "ti", "nz", "*g", "zz", NULL); eG(3);
  printf("listing:\n%s", arg);
  return 0;
eX_2: eX_3:
  eMSG("error! (%s)\n", ePGMs);
}
#endif
//-----------------------------------------------------------------------
